@@ -40,10 +40,6 @@
  */
 #define _XOPEN_SOURCE 700
 
-#if defined(HAVE_ISSETUGID) && defined(__FreeBSD__)
-#define __BSD_VISIBLE 1
-#endif
-
 #include "compat.h"
 #include "debug.h"
 
@@ -270,6 +266,7 @@ p11_dl_error (void)
 {
 	DWORD code = GetLastError();
 	LPVOID msg_buf;
+	char *result;
 
 	FormatMessageA (FORMAT_MESSAGE_ALLOCATE_BUFFER |
 	                FORMAT_MESSAGE_FROM_SYSTEM |
@@ -278,7 +275,9 @@ p11_dl_error (void)
 	                MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
 	                (LPSTR)&msg_buf, 0, NULL);
 
-	return msg_buf;
+	result = strdup (msg_buf);
+	LocalFree (msg_buf);
+	return result;
 }
 
 int
@@ -567,6 +566,7 @@ vasprintf (char **strp,
 	char *nbuf;
 	int guess = 128;
 	int length = 0;
+	va_list orig, aq;
 	int ret;
 
 	if (fmt == NULL) {
@@ -574,17 +574,21 @@ vasprintf (char **strp,
 		return -1;
 	}
 
+	va_copy (orig, ap);
 	for (;;) {
 		nbuf = realloc (buf, guess);
 		if (!nbuf) {
 			free (buf);
+			va_end (orig);
 			return -1;
 		}
 
 		buf = nbuf;
 		length = guess;
 
-		ret = vsnprintf (buf, length, fmt, ap);
+		va_copy (aq, orig);
+		ret = vsnprintf (buf, length, fmt, aq);
+		va_end (aq);
 
 		if (ret < 0)
 			guess *= 2;
@@ -595,6 +599,7 @@ vasprintf (char **strp,
 		else
 			break;
 	}
+	va_end (orig);
 
 	*strp = buf;
 	return ret;
@@ -810,7 +815,8 @@ getauxval (unsigned long type)
 		extern int __libc_enable_secure;
 		secure = __libc_enable_secure;
 
-#elif defined(HAVE_ISSETUGID) && defined(HAVE_ISSETUGID_OPENBSD)
+#elif defined(HAVE_ISSETUGID) && \
+	!((defined __APPLE__ && defined __MACH__) || (defined __FREEBSD__))
 		secure = issetugid ();
 
 #elif defined(OS_UNIX)
