@@ -241,20 +241,14 @@ p11_rpc_message_write_attribute_buffer (p11_rpc_message *msg,
 	/* Write the number of items */
 	p11_rpc_buffer_add_uint32 (msg->output, num);
 
-	/* Is arr not null ? */
-	p11_rpc_buffer_add_byte (msg->output, arr ? 0 : 1);
-
 	for (i = 0; i < num; ++i) {
 		attr = &(arr[i]);
 
 		/* The attribute type */
 		p11_rpc_buffer_add_uint32 (msg->output, attr->type);
 
-		/* rpc_server.c:value_is_null, is the value buffer not NULL ? */
-		p11_rpc_buffer_add_byte (msg->output, attr->pValue ? 0 : 1);
-
 		/* And the attribute buffer length */
-		p11_rpc_buffer_add_uint32 (msg->output, attr->ulValueLen);
+		p11_rpc_buffer_add_uint32 (msg->output, attr->pValue ? attr->ulValueLen : 0);
 	}
 
 	return !p11_buffer_failed (msg->output);
@@ -342,16 +336,14 @@ p11_rpc_message_write_ulong (p11_rpc_message *msg,
 
 bool
 p11_rpc_message_write_byte_buffer (p11_rpc_message *msg,
-                                   CK_BYTE_PTR array,
-                                   CK_ULONG_PTR n_array)
+                                   CK_ULONG count)
 {
 	assert (msg != NULL);
 	assert (msg->output != NULL);
 
 	/* Make sure this is in the right order */
 	assert (!msg->signature || p11_rpc_message_verify_part (msg, "fy"));
-	p11_rpc_buffer_add_byte (msg->output, array ? 0 : 1);
-	p11_rpc_buffer_add_uint32 (msg->output, *n_array);
+	p11_rpc_buffer_add_uint32 (msg->output, count);
 	return !p11_buffer_failed (msg->output);
 }
 
@@ -380,16 +372,14 @@ p11_rpc_message_write_byte_array (p11_rpc_message *msg,
 
 bool
 p11_rpc_message_write_ulong_buffer (p11_rpc_message *msg,
-                                    CK_ULONG_PTR array,
-                                    CK_ULONG_PTR n_array)
+                                    CK_ULONG count)
 {
 	assert (msg != NULL);
 	assert (msg->output != NULL);
 
 	/* Make sure this is in the right order */
 	assert (!msg->signature || p11_rpc_message_verify_part (msg, "fu"));
-	p11_rpc_buffer_add_byte (msg->output, array ? 0 : 1);
-	p11_rpc_buffer_add_uint32 (msg->output, *n_array);
+	p11_rpc_buffer_add_uint32 (msg->output, count);
 	return !p11_buffer_failed (msg->output);
 }
 
